import math
import logging

import torch

import e3nn.o3
import e3nn.nn

from nequip.nn import GraphModuleMixin, GraphModel
from nequip.utils import Config


# == Load old state ==
def initialize_from_state(
    config: Config, graph_model: GraphModel, initialize: bool
) -> GraphModel:
    """Initialize the model from the state dict file given by the config options `initial_model_state`.

    Only loads the state dict if `initialize` is `True`; this is meant for, say, starting a training from a previous state.

    If `initial_model_state_strict` controls
    > whether to strictly enforce that the keys in state_dict
    > match the keys returned by this module's state_dict() function

    See https://pytorch.org/docs/stable/generated/torch.nn.Module.html?highlight=load_state_dict#torch.nn.Module.load_state_dict.
    """
    if not initialize:
        return graph_model  # do nothing
    return load_model_state(
        config=config,
        graph_model=graph_model,
        initialize=initialize,
        _prefix="initial_model_state",
    )


def load_model_state(
    config: Config,
    graph_model: GraphModel,
    initialize: bool,
    _prefix: str = "load_model_state",
) -> GraphModel:
    """Load the model from the state dict file given by the config options `load_model_state`.

    Loads the state dict always; this is meant, for example, for building a new model to deploy with a given state dict.

    If `load_model_state_strict` controls
    > whether to strictly enforce that the keys in state_dict
    > match the keys returned by this module's state_dict() function

    See https://pytorch.org/docs/stable/generated/torch.nn.Module.html?highlight=load_state_dict#torch.nn.Module.load_state_dict.
    """
    if _prefix not in config:
        raise KeyError(
            f"initialize_from_state requires the `{_prefix}` option specifying the state to initialize from"
        )
    # Make sure we map to CPU if there is no GPU, otherwise just leave it alone
    state = torch.load(
        config[_prefix], map_location=None if torch.cuda.is_available() else "cpu"
    )
    strict: bool = config.get(_prefix + "_strict", True)
    graph_model.load_state_dict(state, strict=strict)
    logging.info(
        f"Loaded model state {'' if strict else ' with strict=False'} (parameters/weights/persistent buffers) from state {_prefix}={config[_prefix]}"
    )
    return graph_model


# == Init functions ==
def unit_uniform_init_(t: torch.Tensor):
    """Uniform initialization with <x_i^2> = 1"""
    t.uniform_(-math.sqrt(3), math.sqrt(3))


# TODO: does this normalization make any sense
# def unit_orthogonal_init_(t: torch.Tensor):
#     """Orthogonal init with <x_i^2> = 1"""
#     assert t.ndim == 2
#     torch.nn.init.orthogonal_(t, gain=math.sqrt(max(t.shape)))


def uniform_initialize_FCs(model: GraphModuleMixin, initialize: bool):
    """Initialize ``e3nn.nn.FullyConnectedNet``s with unit uniform initialization"""
    if initialize:

        def _uniform_initialize_fcs(mod: torch.nn.Module):
            if isinstance(mod, e3nn.nn.FullyConnectedNet):
                for layer in mod:
                    # in FC, normalization is expected
                    unit_uniform_init_(layer.weight)

        with torch.no_grad():
            model.apply(_uniform_initialize_fcs)
    return model



from ._one_hot import OneHotAtomEncoding
from ._edge import (
    SphericalHarmonicEdgeAttrs,
    RadialBasisEdgeEncoding,
    AddRadialCutoffToData,
)

__all__ = [
    OneHotAtomEncoding,
    SphericalHarmonicEdgeAttrs,
    RadialBasisEdgeEncoding,
    AddRadialCutoffToData,
]



import tempfile
import functools
import itertools
from os.path import dirname, basename, abspath
from typing import Dict, Any, List, Union, Optional, Sequence

import ase
import ase.io

import torch
import torch.multiprocessing as mp


from nequip.utils.multiprocessing import num_tasks
from .. import AtomicData
from ..transforms import TypeMapper
from ._base_datasets import AtomicInMemoryDataset


def _ase_dataset_reader(
    rank: int,
    world_size: int,
    tmpdir: str,
    ase_kwargs: dict,
    atomicdata_kwargs: dict,
    include_frames,
    global_options: dict,
) -> Union[str, List[AtomicData]]:
    """Parallel reader for all frames in file."""
    if world_size > 1:
        from nequip.utils._global_options import _set_global_options

        # ^ avoid import loop
        # we only `multiprocessing` if world_size > 1
        _set_global_options(global_options)
    # interleave--- in theory it is better for performance for the ranks
    # to read consecutive blocks, but the way ASE is written the whole
    # file gets streamed through all ranks anyway, so just trust the OS
    # to cache things sanely, which it will.
    # ASE handles correctly the case where there are no frames in index
    # and just gives an empty list, so that will succeed:
    index = slice(rank, None, world_size)
    if include_frames is None:
        # count includes 0, 1, ..., inf
        include_frames = itertools.count()

    datas = []
    # stream them from ase too using iread
    for i, atoms in enumerate(ase.io.iread(**ase_kwargs, index=index, parallel=False)):
        global_index = rank + (world_size * i)
        datas.append(
            (
                global_index,
                (
                    AtomicData.from_ase(atoms=atoms, **atomicdata_kwargs)
                    if global_index in include_frames
                    # in-memory dataset will ignore this later, but needed for indexing to work out
                    else None
                ),
            )
        )
    # Save to a tempfile---
    # there can be a _lot_ of tensors here, and rather than dealing with
    # the complications of running out of file descriptors and setting
    # sharing methods, since this is a one time thing, just make it simple
    # and avoid shared memory entirely.
    if world_size > 1:
        path = f"{tmpdir}/rank{rank}.pth"
        torch.save(datas, path)
        return path
    else:
        return datas


class ASEDataset(AtomicInMemoryDataset):
    """

    Args:
        ase_args (dict): arguments for ase.io.read
        include_keys (list): in addition to forces and energy, the keys that needs to
             be parsed into dataset
             The data stored in ase.atoms.Atoms.array has the lowest priority,
             and it will be overrided by data in ase.atoms.Atoms.info
             and ase.atoms.Atoms.calc.results. Optional
        key_mapping (dict): rename some of the keys to the value str. Optional

    Example: Given an atomic data stored in "H2.extxyz" that looks like below:

    ```H2.extxyz
    2
    Properties=species:S:1:pos:R:3 energy=-10 user_label=2.0 pbc="F F F"
     H       0.00000000       0.00000000       0.00000000
     H       0.00000000       0.00000000       1.02000000
    ```

    The yaml input should be

    ```
    dataset: ase
    dataset_file_name: H2.extxyz
    ase_args:
      format: extxyz
    include_keys:
      - user_label
    key_mapping:
      user_label: label0
    chemical_symbols:
      - H
    ```

    for VASP parser, the yaml input should be
    ```
    dataset: ase
    dataset_file_name: OUTCAR
    ase_args:
      format: vasp-out
    key_mapping:
      free_energy: total_energy
    chemical_symbols:
      - H
    ```

    """

    def __init__(
        self,
        root: str,
        ase_args: dict = {},
        file_name: Optional[str] = None,
        url: Optional[str] = None,
        AtomicData_options: Dict[str, Any] = {},
        include_frames: Optional[List[int]] = None,
        type_mapper: TypeMapper = None,
        key_mapping: Optional[dict] = None,
        include_keys: Optional[List[str]] = None,
    ):
        self.ase_args = {}
        self.ase_args.update(getattr(type(self), "ASE_ARGS", dict()))
        self.ase_args.update(ase_args)
        assert "index" not in self.ase_args
        assert "filename" not in self.ase_args

        self.include_keys = include_keys
        self.key_mapping = key_mapping

        super().__init__(
            file_name=file_name,
            url=url,
            root=root,
            AtomicData_options=AtomicData_options,
            include_frames=include_frames,
            type_mapper=type_mapper,
        )

    @classmethod
    def from_atoms_list(cls, atoms: Sequence[ase.Atoms], **kwargs):
        """Make an ``ASEDataset`` from a list of ``ase.Atoms`` objects.

        If `root` is not provided, a temporary directory will be used.

        Please note that this is a convinience method that does NOT avoid a round-trip to disk; the provided ``atoms`` will be written out to a file.

        Ignores ``kwargs["file_name"]`` if it is provided.

        Args:
            atoms
            **kwargs: passed through to the constructor
        Returns:
            The constructed ``ASEDataset``.
        """
        if "root" not in kwargs:
            tmpdir = tempfile.TemporaryDirectory()
            kwargs["root"] = tmpdir.name
        else:
            tmpdir = None
        kwargs["file_name"] = tmpdir.name + "/atoms.xyz"
        atoms = list(atoms)
        # Write them out
        ase.io.write(kwargs["file_name"], atoms, format="extxyz")
        # Read them in
        obj = cls(**kwargs)
        if tmpdir is not None:
            # Make it keep a reference to the tmpdir to keep it alive
            # When the dataset is garbage collected, the tmpdir will
            # be too, and will (hopefully) get deleted eventually.
            # Or at least by end of program...
            obj._tmpdir_ref = tmpdir
        return obj

    @property
    def raw_file_names(self):
        return [basename(self.file_name)]

    @property
    def raw_dir(self):
        return dirname(abspath(self.file_name))

    def get_data(self):
        ase_args = {"filename": self.raw_dir + "/" + self.raw_file_names[0]}
        ase_args.update(self.ase_args)

        # skip the None arguments
        kwargs = dict(
            include_keys=self.include_keys,
            key_mapping=self.key_mapping,
        )
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        kwargs.update(self.AtomicData_options)
        n_proc = num_tasks()
        with tempfile.TemporaryDirectory() as tmpdir:
            from nequip.utils._global_options import _get_latest_global_options

            # ^ avoid import loop
            reader = functools.partial(
                _ase_dataset_reader,
                world_size=n_proc,
                tmpdir=tmpdir,
                ase_kwargs=ase_args,
                atomicdata_kwargs=kwargs,
                include_frames=self.include_frames,
                # get the global options of the parent to initialize the worker correctly
                global_options=_get_latest_global_options(),
            )
            if n_proc > 1:
                # things hang for some obscure OpenMP reason on some systems when using `fork` method
                ctx = mp.get_context("forkserver")
                with ctx.Pool(processes=n_proc) as p:
                    # map it over the `rank` argument
                    datas = p.map(reader, list(range(n_proc)))
                # clean up the pool before loading the data
                datas = [torch.load(d) for d in datas]
                datas = sum(datas, [])
                # un-interleave the datas
                datas = sorted(datas, key=lambda e: e[0])
            else:
                datas = reader(rank=0)
                # datas here is already in order, stride 1 start 0
                # no need to un-interleave
        # return list of AtomicData:
        return [e[1] for e in datas]



from setuptools import setup, find_packages
from pathlib import Path

# see https://packaging.python.org/guides/single-sourcing-package-version/
version_dict = {}
with open(Path(__file__).parents[0] / "nequip/_version.py") as fp:
    exec(fp.read(), version_dict)
version = version_dict["__version__"]
del version_dict

setup(
    name="nequip",
    version=version,
    description="NequIP is an open-source code for building E(3)-equivariant interatomic potentials.",
    download_url="https://github.com/mir-group/nequip",
    author="Simon Batzner, Albert Musealian, Lixin Sun, Anders Johansson, Mario Geiger, Tess Smidt",
    python_requires=">=3.7",
    packages=find_packages(include=["nequip", "nequip.*"]),
    entry_points={
        # make the scripts available as command line scripts
        "console_scripts": [
            "nequip-train = nequip.scripts.train:main",
            "nequip-evaluate = nequip.scripts.evaluate:main",
            "nequip-benchmark = nequip.scripts.benchmark:main",
            "nequip-deploy = nequip.scripts.deploy:main",
        ]
    },
    install_requires=[
        "numpy",
        "ase",
        "tqdm",
        "e3nn>=0.4.4,<0.6.0",
        "pyyaml",
        "contextlib2;python_version<'3.7'",  # backport of nullcontext
        'contextvars;python_version<"3.7"',  # backport of contextvars for savenload
        "typing_extensions;python_version<'3.8'",  # backport of Final
        "importlib_metadata;python_version<'3.10'",  # backport of importlib
        "torch-runstats>=0.2.0",
        "torch-ema>=0.3.0",
    ],
    zip_safe=True,
)



import os

_has_sched_getaffinity: bool = hasattr(os, "sched_getaffinity")


def num_tasks() -> int:
    # sched_getaffinity gives number of _allowed_ cores
    # this is correct for SLURM jobs, for example
    num_avail: int
    if _has_sched_getaffinity:
        num_avail = len(os.sched_getaffinity(0))
    else:
        # on macOS, at least, sched_getaffinity() doesn't appear to be available.
        # fallback to something sane
        num_avail = os.cpu_count()
    # If we couldn't get affinity, don't default to the whole system... sane default to 1
    n_proc: int = int(
        os.environ.get("NEQUIP_NUM_TASKS", num_avail if _has_sched_getaffinity else 1)
    )
    assert n_proc > 0
    assert (
        n_proc <= num_avail
    ), f"Asked for more worker tasks NEQUIP_NUM_TASKS={n_proc} than available CPU cores {num_avail}"
    return n_proc



import warnings
from packaging import version
import os

import torch

import e3nn
import e3nn.util.jit

from .misc import dtype_from_name
from .test import set_irreps_debug
from .config import Config


# for multiprocessing, we need to keep track of our latest global options so
# that we can reload/reset them in worker processes. While we could be more careful here,
# to keep only relevant keys, configs should have only small values (no big objects)
# and those should have references elsewhere anyway, so keeping references here is fine.
_latest_global_config = {}


def _get_latest_global_options() -> dict:
    """Get the config used latest to ``_set_global_options``.

    This is useful for getting worker processes into the same state as the parent.
    """
    global _latest_global_config
    return _latest_global_config


def _set_global_options(config, warn_on_override: bool = False) -> None:
    """Configure global options of libraries like `torch` and `e3nn` based on `config`.

    Args:
        warn_on_override: if True, will try to warn if new options are inconsistant with previously set ones.
    """
    # update these options into the latest global config.
    global _latest_global_config
    _latest_global_config.update(Config.as_dict(config))
    # Set TF32 support
    # See https://pytorch.org/docs/stable/notes/cuda.html#tensorfloat-32-tf32-on-ampere-devices
    if torch.cuda.is_available() and "allow_tf32" in config:
        if torch.torch.backends.cuda.matmul.allow_tf32 is not config["allow_tf32"]:
            # update the setting
            if warn_on_override:
                warnings.warn(
                    f"Setting the GLOBAL value for allow_tf32 to {config['allow_tf32']} which is different than the previous value of {torch.torch.backends.cuda.matmul.allow_tf32}"
                )
            torch.backends.cuda.matmul.allow_tf32 = config["allow_tf32"]
            torch.backends.cudnn.allow_tf32 = config["allow_tf32"]

    # Temporary warning due to unresolved upstream issue
    torch_version = version.parse(torch.__version__)

    if torch_version >= version.parse("1.11"):
        # PyTorch >= 1.11
        k = "_jit_fusion_strategy"
        if k in config:
            new_strat = config.get(k)
            old_strat = torch.jit.set_fusion_strategy(new_strat)
            if warn_on_override and old_strat != new_strat:
                warnings.warn(
                    f"Setting the GLOBAL value for jit fusion strategy to `{new_strat}` which is different than the previous value of `{old_strat}`"
                )
    else:
        # For avoiding 20 steps of painfully slow JIT recompilation
        # See https://github.com/pytorch/pytorch/issues/52286
        k = "_jit_bailout_depth"
        if k in config:
            new_depth = config[k]
            old_depth = torch._C._jit_set_bailout_depth(new_depth)
            if warn_on_override and old_depth != new_depth:
                warnings.warn(
                    f"Setting the GLOBAL value for jit bailout depth to `{new_depth}` which is different than the previous value of `{old_depth}`"
                )

    # Deal with fusers
    # The default PyTorch fuser changed to nvFuser in 1.12
    # fuser1 is NNC, fuser2 is nvFuser
    # See https://github.com/pytorch/pytorch/blob/master/torch/csrc/jit/OVERVIEW.md#fusers
    # And https://github.com/pytorch/pytorch/blob/e0a0f37a11164f59b42bc80a6f95b54f722d47ce/torch/jit/_fuser.py#L46
    # Also https://github.com/pytorch/pytorch/blob/main/torch/csrc/jit/codegen/cuda/README.md
    # Also https://github.com/pytorch/pytorch/blob/66fb83293e6a6f527d3fde632e3547fda20becea/torch/csrc/jit/OVERVIEW.md?plain=1#L1201
    # https://github.com/search?q=repo%3Apytorch%2Fpytorch%20PYTORCH_JIT_USE_NNC_NOT_NVFUSER&type=code
    # We follow the approach they have explicitly built for disabling nvFuser in favor of NNC:
    # https://github.com/pytorch/pytorch/blob/66fb83293e6a6f527d3fde632e3547fda20becea/torch/csrc/jit/codegen/cuda/README.md?plain=1#L214
    #
    #     There are three ways to disable nvfuser. Listed below with descending priorities:
    #      - Force using NNC instead of nvfuser for GPU fusion with env variable `export PYTORCH_JIT_USE_NNC_NOT_NVFUSER=1`.
    #      - Disabling nvfuser with torch API `torch._C._jit_set_nvfuser_enabled(False)`.
    #      - Disable nvfuser with env variable `export PYTORCH_JIT_ENABLE_NVFUSER=0`.
    #
    k = "PYTORCH_JIT_USE_NNC_NOT_NVFUSER"
    if k in os.environ:
        warnings.warn(
            "Do NOT manually set PYTORCH_JIT_USE_NNC_NOT_NVFUSER=0 unless you know exactly what you're doing!"
        )
    else:
        os.environ[k] = "1"

    # TODO: warn_on_override for the rest here?
    if config.get("model_debug_mode", False):
        set_irreps_debug(enabled=True)
    if "default_dtype" in config:
        old_dtype = torch.get_default_dtype()
        new_dtype = dtype_from_name(config["default_dtype"])
        if warn_on_override and old_dtype != new_dtype:
            warnings.warn(
                f"Setting the GLOBAL value for torch.set_default_dtype to `{new_dtype}` which is different than the previous value of `{old_dtype}`"
            )
        torch.set_default_dtype(new_dtype)
    if config.get("grad_anomaly_mode", False):
        torch.autograd.set_detect_anomaly(True)

    e3nn.set_optimization_defaults(**config.get("e3nn_optimization_defaults", {}))

    return



import logging
import secrets

from nequip.utils import Config

import wandb
from wandb.util import json_friendly_val


def init_n_update(config):
    conf_dict = Config.as_dict(config)
    # wandb mangles keys (in terms of type) as well, but we can't easily correct that because there are many ambiguous edge cases. (E.g. string "-1" vs int -1 as keys, are they different config keys?)
    if any(not isinstance(k, str) for k in conf_dict.keys()):
        raise TypeError(
            "Due to wandb limitations, only string keys are supported in configurations."
        )

    # create a run id
    # see https://github.com/wandb/wandb/pull/4676
    config.run_id = secrets.token_urlsafe()

    # download from wandb set up
    wandb.init(
        project=config.wandb_project,
        config=conf_dict,
        name=config.run_name,
        resume="allow",
        id=config.run_id,
    )
    # # download from wandb set up
    updated_parameters = dict(wandb.config)
    for k, v_new in updated_parameters.items():
        skip = False
        if k in config.keys():
            # double check the one sanitized by wandb
            # because we're preprocessing the config and looping over
            # _every_ key, don't mark accessed keys as valid => _get_nomark
            v_old = json_friendly_val(config._get_nomark(k))
            if repr(v_new) == repr(v_old):
                skip = True
        if skip:
            logging.info(f"# skipping wandb update {k} from {v_old} to {v_new}")
        else:
            config.update({k: v_new})
            logging.info(f"# wandb update {k} from {v_old} to {v_new}")
    return config


def resume(config):
    # resume to the old wandb run
    wandb.init(
        project=config.wandb_project,
        resume="must",
        id=config.run_id,
    )



from .auto_init import (
    instantiate_from_cls_name,
    instantiate,
    get_w_prefix,
)
from .savenload import (
    save_file,
    load_file,
    load_callable,
    atomic_write,
    finish_all_writes,
    atomic_write_group,
)
from .config import Config
from .output import Output
from .modules import find_first_of_type
from .misc import dtype_from_name, torch_default_dtype

__all__ = [
    instantiate_from_cls_name,
    instantiate,
    get_w_prefix,
    save_file,
    load_file,
    load_callable,
    atomic_write,
    finish_all_writes,
    atomic_write_group,
    Config,
    Output,
    find_first_of_type,
    dtype_from_name,
    torch_default_dtype,
]



from typing import Optional

from tqdm.auto import tqdm

import torch

from nequip.nn import GraphModel, SequentialGraphNetwork
from nequip.nn import (
    GaussianMixtureModelUncertainty as GaussianMixtureModelUncertaintyModule,
)
from nequip.data import AtomicDataDict, AtomicData, AtomicDataset, Collater
from nequip.utils import find_first_of_type


def GaussianMixtureModelUncertainty(
    graph_model: GraphModel,
    config,
    deploy: bool,
    initialize: bool,
    dataset: Optional[AtomicDataset] = None,
    feature_field: str = AtomicDataDict.NODE_FEATURES_KEY,
    out_field: Optional[str] = None,
):
    r"""Use a GMM on some latent features to predict an uncertainty.

    Only for deployment time!  See `configs/minimal_gmm.yaml`.
    """
    # it only makes sense to add or fit a GMM to a deployment model whose features are already trained
    if (not deploy) or initialize:
        raise RuntimeError(
            "GaussianMixtureModelUncertainty can only be used at deployment time, see `configs/minimal_gmm.yaml`."
        )

    # = add GMM =
    if out_field is None:
        out_field = feature_field + "_nll"

    # TODO: this is VERY brittle!!!!
    seqnn: SequentialGraphNetwork = find_first_of_type(
        graph_model, SequentialGraphNetwork
    )

    gmm: GaussianMixtureModelUncertaintyModule = seqnn.append_from_parameters(
        builder=GaussianMixtureModelUncertaintyModule,
        name=feature_field + "_gmm",
        shared_params=config,
        params=dict(feature_field=feature_field, out_field=out_field),
    )

    if dataset is None:
        raise RuntimeError(
            "GaussianMixtureModelUncertainty requires a dataset to fit the GMM on; did you specify `nequip-deploy --using-dataset`?"
        )

    # = evaluate features =
    # set up model
    prev_training: bool = graph_model.training
    prev_device: torch.device = graph_model.get_device()
    device = config.get("device", None)
    graph_model.eval()
    graph_model.to(device=device)
    # evaluate
    features = []
    collater = Collater.for_dataset(dataset=dataset)
    batch_size: int = config.get("validation_batch_size", config.batch_size)
    stride: int = config.get("dataset_statistics_stride", 1)
    # TODO: guard TQDM on interactive?
    for batch_start_i in tqdm(
        range(0, len(dataset), stride * batch_size),
        desc="GMM eval features on train set",
    ):
        batch = collater(
            [dataset[batch_start_i + i * stride] for i in range(batch_size)]
        )
        # TODO: !! assumption that final value of feature_field is what the
        #          GMM gets is very brittle, should really be extracting it
        #          from the GMM module somehow... not sure how that works.
        #          give it a training mode and exfiltrate it through a buffer?
        #          it is correct, however, for NequIP and Allegro energy models
        features.append(
            graph_model(AtomicData.to_AtomicDataDict(batch.to(device=device)))[
                feature_field
            ]
            .detach()
            .to("cpu")  # offload to not run out of GPU RAM
        )
    features = torch.cat(features, dim=0)
    assert features.ndim == 2
    # restore model
    graph_model.train(mode=prev_training)
    graph_model.to(device=prev_device)
    # fit GMM
    gmm.fit(features, seed=config["seed"])
    del features

    return graph_model



import numpy as np
import logging
import inspect
import itertools
import yaml
import hashlib
import math
from typing import Tuple, Dict, Any, List, Callable, Union, Optional

import torch

from torch_runstats.scatter import scatter_std, scatter_mean

from nequip.utils.torch_geometric import Batch, Dataset
from nequip.utils.torch_geometric.utils import download_url, extract_zip

import nequip
from nequip.data import (
    AtomicData,
    AtomicDataDict,
    _NODE_FIELDS,
    _EDGE_FIELDS,
    _GRAPH_FIELDS,
)
from nequip.utils.batch_ops import bincount
from nequip.utils.regressor import solver
from nequip.utils.savenload import atomic_write
from ..transforms import TypeMapper


class AtomicDataset(Dataset):
    """The base class for all NequIP datasets."""

    root: str
    dtype: torch.dtype

    def __init__(
        self,
        root: str,
        type_mapper: Optional[TypeMapper] = None,
    ):
        self.dtype = torch.get_default_dtype()
        super().__init__(root=root, transform=type_mapper)

    def statistics(
        self,
        fields: List[Union[str, Callable]],
        modes: List[str],
        stride: int = 1,
        unbiased: bool = True,
        kwargs: Optional[Dict[str, dict]] = {},
    ) -> List[tuple]:
        # TODO: If needed, this can eventually be implimented for general AtomicDataset by computing an online running mean and using Welford's method for a stable running standard deviation: https://jonisalonen.com/2013/deriving-welfords-method-for-computing-variance/
        # That would be needed if we have lazy loading datasets.
        # TODO: When lazy-loading datasets are implimented, how to deal with statistics, sampling, and subsets?
        raise NotImplementedError("not implimented for general AtomicDataset yet")

    @property
    def type_mapper(self) -> Optional[TypeMapper]:
        # self.transform is always a TypeMapper
        return self.transform

    def _get_parameters(self) -> Dict[str, Any]:
        """Get a dict of the parameters used to build this dataset."""
        pnames = list(inspect.signature(self.__init__).parameters)
        IGNORE_KEYS = {
            # the type mapper is applied after saving, not before, so doesn't matter to cache validity
            "type_mapper"
        }
        params = {
            k: getattr(self, k)
            for k in pnames
            if k not in IGNORE_KEYS and hasattr(self, k)
        }
        # Add other relevant metadata:
        params["dtype"] = str(self.dtype)
        params["nequip_version"] = nequip.__version__
        return params

    @property
    def processed_dir(self) -> str:
        # We want the file name to change when the parameters change
        # So, first we get all parameters:
        params = self._get_parameters()
        # Make some kind of string of them:
        # we don't care about this possibly changing between python versions,
        # since a change in python version almost certainly means a change in
        # versions of other things too, and is a good reason to recompute
        buffer = yaml.dump(params).encode("ascii")
        # And hash it:
        param_hash = hashlib.sha1(buffer).hexdigest()
        return f"{self.root}/processed_dataset_{param_hash}"


class AtomicInMemoryDataset(AtomicDataset):
    r"""Base class for all datasets that fit in memory.

    Please note that, as a ``pytorch_geometric`` dataset, it must be backed by some kind of disk storage.
    By default, the raw file will be stored at root/raw and the processed torch
    file will be at root/process.

    Subclasses must implement:
     - ``raw_file_names``
     - ``get_data()``

    Subclasses may implement:
     - ``download()`` or ``self.url`` or ``ClassName.URL``

    Args:
        root (str, optional): Root directory where the dataset should be saved. Defaults to current working directory.
        file_name (str, optional): file name of data source. only used in children class
        url (str, optional): url to download data source
        AtomicData_options (dict, optional): extra key that are not stored in data but needed for AtomicData initialization
        include_frames (list, optional): the frames to process with the constructor.
        type_mapper (TypeMapper): the transformation to map atomic information to species index. Optional
    """

    def __init__(
        self,
        root: str,
        file_name: Optional[str] = None,
        url: Optional[str] = None,
        AtomicData_options: Dict[str, Any] = {},
        include_frames: Optional[List[int]] = None,
        type_mapper: Optional[TypeMapper] = None,
    ):
        # TO DO, this may be simplified
        # See if a subclass defines some inputs
        self.file_name = (
            getattr(type(self), "FILE_NAME", None) if file_name is None else file_name
        )
        self.url = getattr(type(self), "URL", url)

        self.AtomicData_options = AtomicData_options
        self.include_frames = include_frames

        self.data = None

        # !!! don't delete this block.
        # otherwise the inherent children class
        # will ignore the download function here
        class_type = type(self)
        if class_type != AtomicInMemoryDataset:
            if "download" not in self.__class__.__dict__:
                class_type.download = AtomicInMemoryDataset.download
            if "process" not in self.__class__.__dict__:
                class_type.process = AtomicInMemoryDataset.process

        # Initialize the InMemoryDataset, which runs download and process
        # See https://pytorch-geometric.readthedocs.io/en/latest/notes/create_dataset.html#creating-in-memory-datasets
        # Then pre-process the data if disk files are not found
        super().__init__(root=root, type_mapper=type_mapper)
        if self.data is None:
            self.data, include_frames = torch.load(self.processed_paths[0])
            if not np.all(include_frames == self.include_frames):
                raise ValueError(
                    f"the include_frames is changed. "
                    f"please delete the processed folder and rerun {self.processed_paths[0]}"
                )

    def len(self):
        if self.data is None:
            return 0
        return self.data.num_graphs

    @property
    def raw_file_names(self):
        raise NotImplementedError()

    @property
    def processed_file_names(self) -> List[str]:
        return ["data.pth", "params.yaml"]

    def get_data(
        self,
    ) -> Union[Tuple[Dict[str, Any], Dict[str, Any]], List[AtomicData]]:
        """Get the data --- called from ``process()``, can assume that ``raw_file_names()`` exist.

        Note that parameters for graph construction such as ``pbc`` and ``r_max`` should be included here as (likely, but not necessarily, fixed) fields.

        Returns:
        A dict:
            fields: dict
                mapping a field name ('pos', 'cell') to a list-like sequence of tensor-like objects giving that field's value for each example.
        Or:
            data_list: List[AtomicData]
        """
        raise NotImplementedError

    def download(self):
        if (not hasattr(self, "url")) or (self.url is None):
            # Don't download, assume present. Later could have FileNotFound if the files don't actually exist
            pass
        else:
            download_path = download_url(self.url, self.raw_dir)
            if download_path.endswith(".zip"):
                extract_zip(download_path, self.raw_dir)

    def process(self):
        data = self.get_data()
        if isinstance(data, list):

            # It's a data list
            data_list = data
            if not (self.include_frames is None or data_list is None):
                data_list = [data_list[i] for i in self.include_frames]
            assert all(isinstance(e, AtomicData) for e in data_list)
            assert all(AtomicDataDict.BATCH_KEY not in e for e in data_list)

            fields = {}

        elif isinstance(data, dict):
            # It's fields
            # Get our data
            fields = data

            # check keys
            all_keys = set(fields.keys())
            assert AtomicDataDict.BATCH_KEY not in all_keys
            # Check bad key combinations, but don't require that this be a graph yet.
            AtomicDataDict.validate_keys(all_keys, graph_required=False)

            # check dimensionality
            num_examples = set([len(a) for a in fields.values()])
            if not len(num_examples) == 1:
                shape_dict = {f: v.shape for f, v in fields.items()}
                raise ValueError(
                    f"This dataset is invalid: expected all fields to have same length (same number of examples), but they had shapes {shape_dict}"
                )
            num_examples = next(iter(num_examples))

            include_frames = self.include_frames
            if include_frames is None:
                include_frames = range(num_examples)

            # Make AtomicData from it:
            if AtomicDataDict.EDGE_INDEX_KEY in all_keys:
                # This is already a graph, just build it
                constructor = AtomicData
            else:
                # do neighborlist from points
                constructor = AtomicData.from_points
                assert "r_max" in self.AtomicData_options
                assert AtomicDataDict.POSITIONS_KEY in all_keys

            data_list = [
                constructor(
                    **{
                        **{f: v[i] for f, v in fields.items()},
                        **self.AtomicData_options,
                    }
                )
                for i in include_frames
            ]

        else:
            raise ValueError("Invalid return from `self.get_data()`")

        # Batch it for efficient saving
        # This limits an AtomicInMemoryDataset to a maximum of LONG_MAX atoms _overall_, but that is a very big number and any dataset that large is probably not "InMemory" anyway
        data = Batch.from_data_list(data_list)
        del data_list
        del fields

        total_MBs = sum(item.numel() * item.element_size() for _, item in data) / (
            1024 * 1024
        )
        logging.info(
            f"Loaded data: {data}\n    processed data size: ~{total_MBs:.2f} MB"
        )
        del total_MBs

        # use atomic writes to avoid race conditions between
        # different trainings that use the same dataset
        # since those separate trainings should all produce the same results,
        # it doesn't matter if they overwrite each others cached'
        # datasets. It only matters that they don't simultaneously try
        # to write the _same_ file, corrupting it.
        with atomic_write(self.processed_paths[0], binary=True) as f:
            torch.save((data, self.include_frames), f)
        with atomic_write(self.processed_paths[1], binary=False) as f:
            yaml.dump(self._get_parameters(), f)

        logging.info("Cached processed data to disk")

        self.data = data

    def get(self, idx):
        return self.data.get_example(idx)

    def _selectors(
        self,
        stride: int = 1,
    ):
        if self._indices is not None:
            graph_selector = torch.as_tensor(self._indices)[::stride]
            # note that self._indices is _not_ necessarily in order,
            # while self.data --- which we take our arrays from ---
            # is always in the original order.
            # In particular, the values of `self.data.batch`
            # are indexes in the ORIGINAL order
            # thus we need graph level properties to also be in the original order
            # so that batch values index into them correctly
            # since self.data.batch is always sorted & contiguous
            # (because of Batch.from_data_list)
            # we sort it:
            graph_selector, _ = torch.sort(graph_selector)
        else:
            graph_selector = torch.arange(0, self.len(), stride)

        node_selector = torch.as_tensor(
            np.in1d(self.data.batch.numpy(), graph_selector.numpy())
        )

        edge_index = self.data[AtomicDataDict.EDGE_INDEX_KEY]
        edge_selector = node_selector[edge_index[0]] & node_selector[edge_index[1]]

        return (graph_selector, node_selector, edge_selector)

    def statistics(
        self,
        fields: List[Union[str, Callable]],
        modes: List[str],
        stride: int = 1,
        unbiased: bool = True,
        kwargs: Optional[Dict[str, dict]] = {},
    ) -> List[tuple]:
        """Compute the statistics of ``fields`` in the dataset.

        If the values at the fields are vectors/multidimensional, they must be of fixed shape and elementwise statistics will be computed.

        Args:
            fields: the names of the fields to compute statistics for.
                Instead of a field name, a callable can also be given that reuturns a quantity to compute the statisics for.

                If a callable is given, it will be called with a (possibly batched) ``Data``-like object and must return a sequence of points to add to the set over which the statistics will be computed.
                The callable must also return a string, one of ``"node"`` or ``"graph"``, indicating whether the value it returns is a per-node or per-graph quantity.
                PLEASE NOTE: the argument to the callable may be "batched", and it may not be batched "contiguously": ``batch`` and ``edge_index`` may have "gaps" in their values.

                For example, to compute the overall statistics of the x,y, and z components of a per-node vector ``force`` field:

                    data.statistics([lambda data: (data.force.flatten(), "node")])

                The above computes the statistics over a set of size 3N, where N is the total number of nodes in the dataset.

            modes: the statistic to compute for each field. Valid options are:
                 - ``count``
                 - ``rms``
                 - ``mean_std``
                 - ``per_atom_*``
                 - ``per_species_*``

            stride: the stride over the dataset while computing statistcs.

            unbiased: whether to use unbiased for standard deviations.

            kwargs: other options for individual statistics modes.

        Returns:
            List of statistics. For fields of floating dtype the statistics are the two-tuple (mean, std); for fields of integer dtype the statistics are a one-tuple (bincounts,)
        """

        # Short circut:
        assert len(modes) == len(fields)
        if len(fields) == 0:
            return []

        graph_selector, node_selector, edge_selector = self._selectors(stride=stride)

        num_graphs = len(graph_selector)
        num_nodes = node_selector.sum()
        num_edges = edge_selector.sum()

        if self.transform is not None:
            # pre-transform the data so that statistics process transformed data
            data_transformed = self.transform(self.data.to_dict(), types_required=False)
        else:
            data_transformed = self.data.to_dict()
        # pre-select arrays
        # this ensures that all following computations use the right data
        all_keys = set()
        selectors = {}
        for k in data_transformed.keys():
            all_keys.add(k)
            if k in _NODE_FIELDS:
                selectors[k] = node_selector
            elif k in _GRAPH_FIELDS:
                selectors[k] = graph_selector
            elif k == AtomicDataDict.EDGE_INDEX_KEY:
                selectors[k] = (slice(None, None, None), edge_selector)
            elif k in _EDGE_FIELDS:
                selectors[k] = edge_selector
        # TODO: do the batch indexes, edge_indexes, etc. after selection need to be
        # "compacted" to subtract out their offsets? For now, we just punt this
        # onto the writer of the callable field.
        # apply selector to actual data
        data_transformed = {
            k: data_transformed[k][selectors[k]]
            for k in data_transformed.keys()
            if k in selectors
        }

        atom_types: Optional[torch.Tensor] = None
        out: list = []
        for ifield, field in enumerate(fields):
            if callable(field):
                # make a joined thing? so it includes fixed fields
                arr, arr_is_per = field(data_transformed)
                arr = arr.to(self.dtype)  # all statistics must be on floating
                assert arr_is_per in ("node", "graph", "edge")
            else:
                if field not in all_keys:
                    raise RuntimeError(
                        f"The field key `{field}` is not present in this dataset"
                    )
                if field not in selectors:
                    # this means field is not selected and so not available
                    raise RuntimeError(
                        f"Only per-node and per-graph fields can have statistics computed; `{field}` has not been registered as either. If it is per-node or per-graph, please register it as such"
                    )
                arr = data_transformed[field]
                if field in _NODE_FIELDS:
                    arr_is_per = "node"
                elif field in _GRAPH_FIELDS:
                    arr_is_per = "graph"
                elif field in _EDGE_FIELDS:
                    arr_is_per = "edge"
                else:
                    raise RuntimeError

            # Check arr
            if arr is None:
                raise ValueError(
                    f"Cannot compute statistics over field `{field}` whose value is None!"
                )
            if not isinstance(arr, torch.Tensor):
                if np.issubdtype(arr.dtype, np.floating):
                    arr = torch.as_tensor(arr, dtype=self.dtype)
                else:
                    arr = torch.as_tensor(arr)
            if arr_is_per == "node":
                arr = arr.view(num_nodes, -1)
            elif arr_is_per == "graph":
                arr = arr.view(num_graphs, -1)
            elif arr_is_per == "edge":
                arr = arr.view(num_edges, -1)

            ana_mode = modes[ifield]
            # compute statistics
            if ana_mode == "count":
                # count integers
                uniq, counts = torch.unique(
                    torch.flatten(arr), return_counts=True, sorted=True
                )
                out.append((uniq, counts))
            elif ana_mode == "rms":
                # root-mean-square
                out.append((torch.sqrt(torch.mean(arr * arr)),))

            elif ana_mode == "mean_std":
                # mean and std
                if len(arr) < 2:
                    raise ValueError(
                        "Can't do per species standard deviation without at least two samples"
                    )
                mean = torch.mean(arr, dim=0)
                std = torch.std(arr, dim=0, unbiased=unbiased)
                out.append((mean, std))

            elif ana_mode == "absmax":
                out.append((arr.abs().max(),))

            elif ana_mode.startswith("per_species_"):
                # per-species
                algorithm_kwargs = kwargs.pop(field + ana_mode, {})

                ana_mode = ana_mode[len("per_species_") :]

                if atom_types is None:
                    atom_types = data_transformed[AtomicDataDict.ATOM_TYPE_KEY]

                results = self._per_species_statistics(
                    ana_mode,
                    arr,
                    arr_is_per=arr_is_per,
                    batch=data_transformed[AtomicDataDict.BATCH_KEY],
                    atom_types=atom_types,
                    unbiased=unbiased,
                    algorithm_kwargs=algorithm_kwargs,
                )
                out.append(results)

            elif ana_mode.startswith("per_atom_"):
                # per-atom
                # only makes sense for a per-graph quantity
                if arr_is_per != "graph":
                    raise ValueError(
                        f"It doesn't make sense to ask for `{ana_mode}` since `{field}` is not per-graph"
                    )
                ana_mode = ana_mode[len("per_atom_") :]
                results = self._per_atom_statistics(
                    ana_mode=ana_mode,
                    arr=arr,
                    batch=data_transformed[AtomicDataDict.BATCH_KEY],
                    unbiased=unbiased,
                )
                out.append(results)

            else:
                raise NotImplementedError(f"Cannot handle statistics mode {ana_mode}")

        return out

    @staticmethod
    def _per_atom_statistics(
        ana_mode: str,
        arr: torch.Tensor,
        batch: torch.Tensor,
        unbiased: bool = True,
    ):
        """Compute "per-atom" statistics that are normalized by the number of atoms in the system.

        Only makes sense for a graph-level quantity (checked by .statistics).
        """
        # using unique_consecutive handles the non-contiguous selected batch index
        _, N = torch.unique_consecutive(batch, return_counts=True)
        N = N.unsqueeze(-1)
        assert N.ndim == 2
        assert N.shape == (len(arr), 1)
        assert arr.ndim >= 2
        data_dim = arr.shape[1:]
        arr = arr / N
        assert arr.shape == (len(N),) + data_dim
        if ana_mode == "mean_std":
            if len(arr) < 2:
                raise ValueError(
                    "Can't do standard deviation without at least two samples"
                )
            mean = torch.mean(arr, dim=0)
            std = torch.std(arr, unbiased=unbiased, dim=0)
            return mean, std
        elif ana_mode == "rms":
            return (torch.sqrt(torch.mean(arr.square())),)
        elif ana_mode == "absmax":
            return (torch.max(arr.abs()),)
        else:
            raise NotImplementedError(
                f"{ana_mode} for per-atom analysis is not implemented"
            )

    def _per_species_statistics(
        self,
        ana_mode: str,
        arr: torch.Tensor,
        arr_is_per: str,
        atom_types: torch.Tensor,
        batch: torch.Tensor,
        unbiased: bool = True,
        algorithm_kwargs: Optional[dict] = {},
    ):
        """Compute "per-species" statistics.

        For a graph-level quantity, models it as a linear combintation of the number of atoms of different types in the graph.

        For a per-node quantity, computes the expected statistic but for each type instead of over all nodes.
        """
        N = bincount(atom_types.squeeze(-1), batch)
        assert N.ndim == 2  # [batch, n_type]
        N = N[(N > 0).any(dim=1)]  # deal with non-contiguous batch indexes
        assert arr.ndim >= 2
        if arr_is_per == "graph":

            if ana_mode != "mean_std":
                raise NotImplementedError(
                    f"{ana_mode} for per species analysis is not implemented for shape {arr.shape}"
                )

            N = N.type(self.dtype)

            return solver(N, arr, **algorithm_kwargs)

        elif arr_is_per == "node":
            arr = arr.type(self.dtype)

            if ana_mode == "mean_std":
                # There need to be at least two occurances of each atom type in the
                # WHOLE dataset, not in any given frame:
                if torch.any(N.sum(dim=0) < 2):
                    raise ValueError(
                        "Can't do per species standard deviation without at least two samples per species"
                    )
                mean = scatter_mean(arr, atom_types, dim=0)
                assert mean.shape[1:] == arr.shape[1:]  # [N, dims] -> [type, dims]
                assert len(mean) == N.shape[1]
                std = scatter_std(arr, atom_types, dim=0, unbiased=unbiased)
                assert std.shape == mean.shape
                return mean, std
            elif ana_mode == "rms":
                square = scatter_mean(arr.square(), atom_types, dim=0)
                assert square.shape[1:] == arr.shape[1:]  # [N, dims] -> [type, dims]
                assert len(square) == N.shape[1]
                dims = len(square.shape) - 1
                for i in range(dims):
                    square = square.mean(axis=-1)
                return (torch.sqrt(square),)
            else:
                raise NotImplementedError(
                    f"Statistics mode {ana_mode} isn't yet implemented for per_species_"
                )

        else:
            raise NotImplementedError

    def rdf(
        self, bin_width: float, stride: int = 1
    ) -> Dict[Tuple[int, int], Tuple[np.ndarray, np.ndarray]]:
        """Compute the pairwise RDFs of the dataset.

        Args:
            bin_width: width of the histogram bin in distance units
            stride: stride of data to include

        Returns:
            dictionary mapping `(type1, type2)` to tuples of `(hist, bin_edges)` in the style of `np.histogram`.
        """
        graph_selector, node_selector, edge_selector = self._selectors(stride=stride)

        data = AtomicData.to_AtomicDataDict(self.data)
        data = AtomicDataDict.with_edge_vectors(data, with_lengths=True)

        results = {}

        types = self.type_mapper(data)[AtomicDataDict.ATOM_TYPE_KEY]

        edge_types = torch.index_select(
            types, 0, data[AtomicDataDict.EDGE_INDEX_KEY].reshape(-1)
        ).view(2, -1)
        types_center = edge_types[0].numpy()
        types_neigh = edge_types[1].numpy()

        r_max: float = self.AtomicData_options["r_max"]
        # + 1 to always have a zero bin at the end
        n_bins: int = int(math.ceil(r_max / bin_width)) + 1
        # +1 since these are bin_edges including rightmost
        bins = bin_width * np.arange(n_bins + 1)

        for type1, type2 in itertools.combinations_with_replacement(
            range(self.type_mapper.num_types), 2
        ):
            # Try to do as much of this as possible in-place
            mask = types_center == type1
            np.logical_and(mask, types_neigh == type2, out=mask)
            np.logical_and(mask, edge_selector, out=mask)
            mask = mask.astype(np.int32)
            results[(type1, type2)] = np.histogram(
                data[AtomicDataDict.EDGE_LENGTH_KEY],
                weights=mask,
                bins=bins,
                density=True,
            )
            # RDF is symmetric
            results[(type2, type1)] = results[(type1, type2)]

        return results



import inspect
from typing import Optional

import torch

from nequip.data import AtomicDataset
from nequip.data.transforms import TypeMapper
from nequip.nn import GraphModuleMixin, GraphModel
from nequip.utils import (
    load_callable,
    instantiate,
    dtype_from_name,
    torch_default_dtype,
    Config,
)
from nequip.utils.config import _GLOBAL_ALL_ASKED_FOR_KEYS


def model_from_config(
    config: Config,
    initialize: bool = False,
    dataset: Optional[AtomicDataset] = None,
    deploy: bool = False,
) -> GraphModuleMixin:
    """Build a model based on `config`.

    Model builders (`model_builders`) can have arguments:
     - ``config``: the config. Always present.
     - ``model``: the model produced by the previous builder. Cannot be requested by the first builder, must be requested by subsequent ones.
     - ``initialize``: whether to initialize the model
     - ``dataset``: if ``initialize`` is True, the dataset
     - ``deploy``: whether the model object is for deployment / inference

    Note that this function temporarily sets ``torch.set_default_dtype()`` and as such is not thread safe.

    Args:
        config
        initialize (bool): whether ``model_builders`` should be instructed to initialize the model
        dataset: dataset for initializers if ``initialize`` is True.
        deploy (bool): whether ``model_builders`` should be told the model is for deployment / inference

    Returns:
        The build model.
    """
    if isinstance(config, dict):
        config = Config.from_dict(config)
    # Pre-process config
    type_mapper = None
    if dataset is not None:
        type_mapper = dataset.type_mapper
    else:
        try:
            type_mapper, _ = instantiate(TypeMapper, all_args=config)
        except RuntimeError:
            pass

    if type_mapper is not None:
        if "num_types" in config:
            assert (
                config["num_types"] == type_mapper.num_types
            ), "inconsistant config & dataset"
        if "type_names" in config:
            assert (
                config["type_names"] == type_mapper.type_names
            ), "inconsistant config & dataset"
        config["num_types"] = type_mapper.num_types
        config["type_names"] = type_mapper.type_names
        config["type_to_chemical_symbol"] = type_mapper.type_to_chemical_symbol
        # We added them, so they are by definition valid:
        _GLOBAL_ALL_ASKED_FOR_KEYS.update(
            {"num_types", "type_names", "type_to_chemical_symbol"}
        )

    default_dtype = torch.get_default_dtype()
    model_dtype: torch.dtype = dtype_from_name(config.get("model_dtype", default_dtype))
    config["model_dtype"] = str(model_dtype).lstrip("torch.")
    # confirm sanity
    assert default_dtype in (torch.float32, torch.float64)
    if default_dtype == torch.float32 and model_dtype == torch.float64:
        raise ValueError(
            "Overall default_dtype=float32, but model_dtype=float64 is a higher precision- change default_dtype to float64"
        )
    # temporarily set the default dtype
    start_graph_model_builders = None
    with torch_default_dtype(model_dtype):

        # Build
        builders = [
            load_callable(b, prefix="nequip.model")
            for b in config.get("model_builders", [])
        ]

        model = None

        for builder_i, builder in enumerate(builders):
            pnames = inspect.signature(builder).parameters
            params = {}
            if "graph_model" in pnames:
                # start graph_model builders, which happen later
                start_graph_model_builders = builder_i
                break
            if "initialize" in pnames:
                params["initialize"] = initialize
            if "deploy" in pnames:
                params["deploy"] = deploy
            if "config" in pnames:
                params["config"] = config
            if "dataset" in pnames:
                if "initialize" not in pnames:
                    raise ValueError(
                        "Cannot request dataset without requesting initialize"
                    )
                if (
                    initialize
                    and pnames["dataset"].default == inspect.Parameter.empty
                    and dataset is None
                ):
                    raise RuntimeError(
                        f"Builder {builder.__name__} requires the dataset, initialize is true, but no dataset was provided to `model_from_config`."
                    )
                params["dataset"] = dataset
            if "model" in pnames:
                if model is None:
                    raise RuntimeError(
                        f"Builder {builder.__name__} asked for the model as an input, but no previous builder has returned a model"
                    )
                params["model"] = model
            else:
                if model is not None:
                    raise RuntimeError(
                        f"All model_builders after the first one that returns a model must take the model as an argument; {builder.__name__} doesn't"
                    )
            model = builder(**params)
            if model is not None and not isinstance(model, GraphModuleMixin):
                raise TypeError(
                    f"Builder {builder.__name__} didn't return a GraphModuleMixin, got {type(model)} instead"
                )
    # reset to default dtype by context manager

    # Wrap the model up
    model = GraphModel(
        model,
        model_dtype=model_dtype,
        model_input_fields=config.get("model_input_fields", {}),
    )

    # Run GraphModel builders
    if start_graph_model_builders is not None:
        for builder in builders[start_graph_model_builders:]:
            pnames = inspect.signature(builder).parameters
            params = {}
            assert "graph_model" in pnames
            params["graph_model"] = model
            if "model" in pnames:
                raise ValueError(
                    f"Once any builder requests `graph_model` (first requested by {builders[start_graph_model_builders].__name__}), no builder can request `model`, but {builder.__name__} did"
                )
            if "initialize" in pnames:
                params["initialize"] = initialize
            if "deploy" in pnames:
                params["deploy"] = deploy
            if "config" in pnames:
                params["config"] = config
            if "dataset" in pnames:
                if "initialize" not in pnames:
                    raise ValueError(
                        "Cannot request dataset without requesting initialize"
                    )
                if (
                    initialize
                    and pnames["dataset"].default == inspect.Parameter.empty
                    and dataset is None
                ):
                    raise RuntimeError(
                        f"Builder {builder.__name__} requires the dataset, initialize is true, but no dataset was provided to `model_from_config`."
                    )
                params["dataset"] = dataset

            model = builder(**params)
            if not isinstance(model, GraphModel):
                raise TypeError(
                    f"Builder {builder.__name__} didn't return a GraphModel, got {type(model)} instead"
                )

    return model



import logging
from typing import List, Optional, Union

import torch

from nequip.nn import RescaleOutput, GraphModuleMixin, PerSpeciesScaleShift
from nequip.data import AtomicDataDict, AtomicDataset
from nequip.data.transforms import TypeMapper


RESCALE_THRESHOLD = 1e-6


def RescaleEnergyEtc(
    model: GraphModuleMixin,
    config,
    initialize: bool,
    dataset: Optional[AtomicDataset] = None,
):
    return GlobalRescale(
        model=model,
        config=config,
        dataset=dataset,
        initialize=initialize,
        module_prefix="global_rescale",
        default_scale=(
            f"dataset_{AtomicDataDict.FORCE_KEY}_rms"
            if AtomicDataDict.FORCE_KEY in model.irreps_out
            else f"dataset_{AtomicDataDict.TOTAL_ENERGY_KEY}_std"
        ),
        default_shift=None,
        default_scale_keys=AtomicDataDict.ALL_ENERGY_KEYS,
        default_shift_keys=[AtomicDataDict.TOTAL_ENERGY_KEY],
    )


def GlobalRescale(
    model: GraphModuleMixin,
    config,
    initialize: bool,
    module_prefix: str,
    default_scale: Union[str, float, list],
    default_shift: Union[str, float, list],
    default_scale_keys: list,
    default_shift_keys: list,
    dataset: Optional[AtomicDataset] = None,
):
    """Add global rescaling for energy(-based quantities).

    If ``initialize`` is false, doesn't compute statistics.
    """

    global_scale = config.get(f"{module_prefix}_scale", default_scale)
    global_shift = config.get(f"{module_prefix}_shift", default_shift)

    if global_shift is not None:
        logging.warning(
            f"!!!! Careful global_shift is set to {global_shift}."
            f"The model for {default_shift_keys} will no longer be size extensive"
        )

    # = Get statistics of training dataset =
    if initialize:
        str_names = []
        for value in [global_scale, global_shift]:
            if isinstance(value, str):
                str_names += [value]
            elif (
                value is None
                or isinstance(value, float)
                or isinstance(value, torch.Tensor)
            ):
                # valid values
                pass
            else:
                raise ValueError(f"Invalid global scale `{value}`")

        # = Compute shifts and scales =
        if len(str_names) > 0:
            computed_stats = _compute_stats(
                str_names=str_names,
                dataset=dataset,
                stride=config.dataset_statistics_stride,
            )

        if isinstance(global_scale, str):
            s = global_scale
            global_scale = computed_stats[str_names.index(global_scale)]
            logging.info(f"Replace string {s} to {global_scale}")
        if isinstance(global_shift, str):
            s = global_shift
            global_shift = computed_stats[str_names.index(global_shift)]
            logging.info(f"Replace string {s} to {global_shift}")

        if global_scale is not None and global_scale < RESCALE_THRESHOLD:
            raise ValueError(
                f"Global energy scaling was very low: {global_scale}. If dataset values were used, does the dataset contain insufficient variation? Maybe try disabling global scaling with global_scale=None."
            )

        logging.info(
            f"Initially outputs are globally scaled by: {global_scale}, total_energy are globally shifted by {global_shift}."
        )

    else:
        # Put dummy values
        if global_shift is not None:
            global_shift = 0.0  # it has some kind of value
        if global_scale is not None:
            global_scale = 1.0  # same,

    error_string = "keys need to be a list"
    assert isinstance(default_scale_keys, list), error_string
    assert isinstance(default_shift_keys, list), error_string

    # == Build the model ==
    return RescaleOutput(
        model=model,
        scale_keys=[k for k in default_scale_keys if k in model.irreps_out],
        scale_by=global_scale,
        shift_keys=[k for k in default_shift_keys if k in model.irreps_out],
        shift_by=global_shift,
        shift_trainable=config.get(f"{module_prefix}_shift_trainable", False),
        scale_trainable=config.get(f"{module_prefix}_scale_trainable", False),
        default_dtype=config.get("default_dtype", None),
    )


def PerSpeciesRescale(
    model: GraphModuleMixin,
    config,
    initialize: bool,
    dataset: Optional[AtomicDataset] = None,
):
    """Add per-atom rescaling (and shifting) for per-atom energies."""
    module_prefix = "per_species_rescale"

    # Check for common double shift mistake with defaults
    if "RescaleEnergyEtc" in config.get("model_builders", []):
        # if the defaults are enabled, then we will get bad double shift
        # THIS CHECK IS ONLY GOOD ENOUGH FOR EMITTING WARNINGS
        has_global_shift = config.get("global_rescale_shift", None) is not None
        if has_global_shift:
            if config.get(module_prefix + "_shifts", True) is not None:
                # using default of per_atom shift
                raise RuntimeError(
                    "A global_rescale_shift was provided, but the default per-atom energy shift was not disabled."
                )
        del has_global_shift

    return _PerSpeciesRescale(
        scales_default=None,
        shifts_default=f"dataset_per_atom_{AtomicDataDict.TOTAL_ENERGY_KEY}_mean",
        field=AtomicDataDict.PER_ATOM_ENERGY_KEY,
        out_field=AtomicDataDict.PER_ATOM_ENERGY_KEY,
        module_prefix=module_prefix,
        insert_before="total_energy_sum",
        model=model,
        config=config,
        initialize=initialize,
        dataset=dataset,
    )


def _PerSpeciesRescale(
    scales_default,
    shifts_default,
    field: str,
    out_field: str,
    module_prefix: str,
    insert_before: str,
    model: GraphModuleMixin,
    config,
    initialize: bool,
    dataset: Optional[AtomicDataset] = None,
):
    """Add per-atom rescaling (and shifting) for a field

    If ``initialize`` is false, doesn't compute statistics.
    """
    scales = config.get(module_prefix + "_scales", scales_default)
    shifts = config.get(module_prefix + "_shifts", shifts_default)

    # = Determine what statistics need to be compute =
    assert config.get(
        module_prefix + "_arguments_in_dataset_units", True
    ), f"The PerSpeciesRescale builder is only compatible with {module_prefix + '_arguments_in_dataset_units'} set to True"

    if initialize:
        str_names = []
        for value in [scales, shifts]:
            if isinstance(value, str):
                str_names += [value]
            elif (
                value is None
                or isinstance(value, float)
                or isinstance(value, list)
                or isinstance(value, torch.Tensor)
            ):
                # valid values
                pass
            else:
                raise ValueError(f"Invalid value `{value}` of type {type(value)}")

        # = Compute shifts and scales =
        if len(str_names) > 0:
            computed_stats = _compute_stats(
                str_names=str_names,
                dataset=dataset,
                stride=config.dataset_statistics_stride,
                kwargs=config.get(module_prefix + "_kwargs", {}),
            )

        if isinstance(scales, str):
            s = scales
            # energy or other property is 1D:
            scales = computed_stats[str_names.index(scales)].squeeze(-1)
            logging.info(f"Replace string {s} to {scales}")
        elif isinstance(scales, (list, float)):
            scales = torch.as_tensor(scales)

        if isinstance(shifts, str):
            s = shifts
            # energy or other property is 1D:
            shifts = computed_stats[str_names.index(shifts)].squeeze(-1)
            logging.info(f"Replace string {s} to {shifts}")
        elif isinstance(shifts, (list, float)):
            shifts = torch.as_tensor(shifts)

        # TODO kind of weird error to check for here
        if scales is not None and torch.min(scales) < RESCALE_THRESHOLD:
            raise ValueError(
                f"Per species scaling was very low: {scales}. Maybe try setting {module_prefix}_scales = 1."
            )

        logging.info(
            f"Atomic outputs are scaled by: {TypeMapper.format(scales, config.type_names)}, shifted by {TypeMapper.format(shifts, config.type_names)}."
        )

    else:
        # Put dummy values
        # the real ones will be loaded from the state dict later
        # note that the state dict includes buffers,
        # so this is fine regardless of whether its trainable.
        scales = 1.0 if scales is not None else None
        shifts = 0.0 if shifts is not None else None
        # values from the previously initialized model
        # will be brought in from the state dict later,
        # so these values (and rescaling them) doesn't matter

    # insert in per species shift
    params = dict(
        field=field,
        out_field=out_field,
        shifts=shifts,
        scales=scales,
        arguments_in_dataset_units=True,
    )
    model.insert_from_parameters(
        before=insert_before,
        name=module_prefix,
        shared_params=config,
        builder=PerSpeciesScaleShift,
        params=params,
    )

    # == Build the model ==
    return model


def _compute_stats(
    str_names: List[str], dataset, stride: int, kwargs: Optional[dict] = {}
):
    """return the values of statistics over dataset
    quantity name should be dataset_key_stat, where key can be any key
    that exists in the dataset, stat can be mean, std

    Args:

    str_names: list of strings that define the quantity to compute
    dataset: dataset object to run the stats over
    stride: # frames to skip for every one frame to include
    """

    # parse the list of string to field, mode
    # and record which quantity correspond to which computed_item
    stat_modes = []
    stat_fields = []
    stat_strs = []
    ids = []
    tuple_ids = []
    tuple_id_map = {"mean": 0, "std": 1, "rms": 0, "absmax": 0}
    input_kwargs = {}
    for name in str_names:

        # remove dataset prefix
        if name.startswith("dataset_"):
            name = name[len("dataset_") :]
        # identify per_species and per_atom modes
        prefix = ""
        if name.startswith("per_species_"):
            name = name[len("per_species_") :]
            prefix = "per_species_"
        elif name.startswith("per_atom_"):
            name = name[len("per_atom_") :]
            prefix = "per_atom_"

        stat = name.split("_")[-1]
        field = "_".join(name.split("_")[:-1])
        if stat in ["mean", "std"]:
            stat_mode = prefix + "mean_std"
            stat_str = field + prefix + "mean_std"
        elif stat in ["rms", "absmax"]:
            stat_mode = prefix + stat
            stat_str = field + prefix + stat
        else:
            raise ValueError(f"Cannot handle {stat} type quantity")

        if stat_str in stat_strs:
            ids += [stat_strs.index(stat_str)]
        else:
            ids += [len(stat_strs)]
            stat_strs += [stat_str]
            stat_modes += [stat_mode]
            stat_fields += [field]
            if stat_mode.startswith("per_species_"):
                if field in kwargs:
                    input_kwargs[field + stat_mode] = kwargs[field]
        tuple_ids += [tuple_id_map[stat]]

    values = dataset.statistics(
        fields=stat_fields,
        modes=stat_modes,
        stride=stride,
        kwargs=input_kwargs,
    )
    return [values[idx][tuple_ids[i]] for i, idx in enumerate(ids)]
